import os
import yaml

try:
    from yaml import CLoader as base_loader, CDumper as Dumper
except ImportError:
    from yaml import Loader as base_loader, Dumper

"""
Extend the loader class to be able to reference files in .yaml files.
Adapted from https://stackoverflow.com/questions/528281/how-can-i-include-a-yaml-file-inside-another
Accessed 2020-07-28
"""


class Loader(base_loader):
    def __init__(self, stream):
        self._root = os.path.split(stream.name)[0]
        super().__init__(stream)

    def scalar_from_file(self, node):
        filename = os.path.join(self._root, self.construct_scalar(node))
        with open(filename, "r") as f:
            return f.read()

    def include_other_yml(self, node):
        filename = os.path.join(self._root, self.construct_scalar(node))
        with open(filename, "r") as f:
            return yaml.load(f, Loader)


Loader.add_constructor("!text_from_file", Loader.scalar_from_file)
Loader.add_constructor("!yml_from_file", Loader.include_other_yml)



from airflow.utils.log.logging_mixin import LoggingMixin

from ewah.hooks.base import EWAHBaseHook
from ewah.constants import EWAHConstants as EC

import json
import bson
import math
import hashlib

from copy import deepcopy
from collections import OrderedDict
from psycopg2.extras import RealDictCursor
from bson.json_util import dumps
from decimal import Decimal
from typing import Optional, Type, Union, Dict


class EWAHJSONEncoder(json.JSONEncoder):
    """Extension of the native json encoder to deal with additional datatypes and
    issues relating to (+/-) Inf and NaN numbers.
    """

    def default(self, obj):
        """Method is called if an object cannot be serialized.

        Ought to return a serializeable value for the object. Ought to raise an error
        if unable to do so.

        Implemented types:
            - Decimal -> float
            - bson.objectid.ObjectId -> string
        """
        if isinstance(obj, bson.objectid.ObjectId):
            # MongoDB Object IDs - return just the ID itself as string
            return str(obj)
        if isinstance(obj, Decimal):
            return float(obj)
        # Let the base class default method raise the TypeError
        return super().default(obj)

    def iterencode(self, o, _one_shot=False):
        """Overwrite the iterencode method because this is where the float
        special cases are handled in the floatstr() function. Copy-pasted
        original code and then adapted it to change floatstr() behavior.

        This was necessary because the json module accepts (+/-) Infinity and NaN
        objects as floats, but PostgreSQL uses the tighter JSON standard and does not
        accept them in json. The json module offers no optional flag to deal with this
        issue natively. Thus, overwrite the iterencode method and remove Inf and NaN
        by returning null instead.
        """
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = json.encoder.encode_basestring_ascii
        else:
            _encoder = json.encoder.encode_basestring

        def floatstr(
            o,
            allow_nan=self.allow_nan,
            _repr=float.__repr__,
            _inf=float("inf"),
            _neginf=-float("inf"),
        ):
            # Check for specials.  Note that this type of test is processor
            # and/or platform-specific, so do tests which don't depend on the
            # internals.
            if not (o != o or o == _inf or o == _neginf):
                return _repr(o)
            if not allow_nan:
                raise ValueError(
                    "Out of range float values are not JSON compliant: " + repr(o)
                )
            return "null"

        return json.encoder._make_iterencode(
            markers,
            self.default,
            _encoder,
            self.indent,
            floatstr,
            self.key_separator,
            self.item_separator,
            self.sort_keys,
            self.skipkeys,
            _one_shot,
        )(o, 0)


class EWAHBaseUploader(LoggingMixin):
    """Base class for all EWAH uploader classes aka Uploaders.

    Uploaders are classes that can receive data in a standardized format and upload
    it to a target data storage. Each Uploader is uploads data to one type of target.
    For instance, the EWAHPostgresUploader loads data to PostgreSQL databases.

    The base class contains the basic methods shared by all Uploaders for the purpose of
    uploading data. Each target shall have a subclass derived from this class.
    """

    def __init__(
        self, dwh_engine: str, dwh_conn: Type[EWAHBaseHook], *args, **kwarg
    ) -> None:
        super().__init__(*args, **kwarg)
        self.dwh_engine = dwh_engine
        self.dwh_conn = dwh_conn
        self.dwh_hook = dwh_conn.get_hook()

    def _get_column_type(self, column_definition: dict) -> str:
        """Return the column type of the field. Returns the DWH engine specific default
        if there is none.

        :param column_definition: Dictionary containing details on this column,
            including optionally the column type.
        """
        return column_definition.get(
            EC.QBC_FIELD_TYPE,
            EC.QBC_TYPE_MAPPING[self.dwh_engine].get(
                EC.QBC_TYPE_MAPPING_DEFAULT,
                EC.QBC_TYPE_MAPPING_INCONSISTENT,
            ),
        )

    def close(self):
        self.dwh_hook.close()

    def copy_table(
        self,
        old_schema: str,
        old_table: str,
        new_schema: str,
        new_table: str,
        **kwargs: Optional[Dict[str, str]],  # e.g. database_name
    ) -> None:
        """Copy a table, including all data.

        :param old_schema: Schema name of the source table.
        :param old_table: Table name of the source table.
        :param new_schema: Schema name of the newly created table.
        :param new_table: Table name of the newly created table.

        :Keyword Arguments:
            * Any argument that may be given as format keyword argument to format the
                ``_COPY_TABLE`` default SQL, e.g. ``database_name`` for Snowflake
        """
        test_kwargs = {"table_name": old_table, "schema_name": old_schema}
        test_kwargs.update(
            {key: value for (key, value) in kwargs.items() if not value is None}
        )
        if self.test_if_table_exists(**test_kwargs):
            self.dwh_hook.execute(
                sql=self._COPY_TABLE.format(
                    old_schema=old_schema,
                    old_table=old_table,
                    new_schema=new_schema,
                    new_table=new_table,
                    **kwargs,
                ),
                commit=False,
            )

    def detect_and_apply_schema_changes(
        self,
        new_schema_name,
        new_table_name,
        new_columns_dictionary,
        drop_missing_columns=False,
        database=None,
        commit=False,
    ):
        params = {
            "schema_name": new_schema_name,
            "table_name": new_table_name,
        }
        if self.dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            database = database or self.database
            params["database_name"] = database
        if not self.test_if_table_exists(**params):
            return ([], [])  # Table did not previously exist, so there is nothing to do

        list_of_old_columns = [
            col[0].strip()
            for col in self.dwh_hook.execute_and_return_result(
                sql=self._QUERY_SCHEMA_CHANGES_COLUMNS.format(**params),
                params=params,
                return_dict=False,
            )
        ]
        list_of_columns = [
            col_name.strip() for col_name in list(new_columns_dictionary.keys())
        ]

        dropped_columns = []
        if drop_missing_columns:
            for old_column in list_of_old_columns:
                if not (old_column in list_of_columns):
                    drop_params = deepcopy(params)
                    drop_params["column_name"] = old_column
                    dropped_columns += [old_column]
                    self.dwh_hook.execute(
                        sql=self._QUERY_SCHEMA_CHANGES_DROP_COLUMN.format(
                            **drop_params,
                        ),
                        commit=False,
                    )

        new_columns = []
        for column in list_of_columns:
            if not (column in list_of_old_columns):
                new_columns += [column]
                add_params = deepcopy(params)
                add_params["column_name"] = column
                add_params["column_type"] = self._get_column_type(
                    new_columns_dictionary[column]
                )
                self.dwh_hook.execute(
                    sql=self._QUERY_SCHEMA_CHANGES_ADD_COLUMN.format(
                        **add_params,
                    ),
                    commit=False,
                )

        if commit:
            self.dwh_hook.commit()

        return (new_columns, dropped_columns)

    def create_or_update_table(
        self,
        data,
        columns_definition,
        table_name,
        schema_name,
        schema_suffix,
        database_name=None,
        drop_and_replace=True,
        update_on_columns=None,
        commit=False,
        clean_data_before_upload=True,
        hash_columns=None,
        hashlib_func_name=None,
    ):
        # check this again with Snowflake!!
        database_name = database_name or getattr(self, "database", None)

        mapped_types = EC.QBC_TYPE_MAPPING[self.dwh_engine].keys()
        mapped_types = tuple([k for k in mapped_types if isinstance(k, type)])

        raw_row = {}  # Used as template for params at execution
        sql_part_columns = []  # Used for CREATE and INSERT / UPDATE query
        jsonb_columns = []  # JSON columns require special treatment
        create_update_on_columns = not (drop_and_replace or update_on_columns)
        update_on_columns = update_on_columns or []
        if not (type([]) == type(update_on_columns)):
            raise Exception(
                '"update_on_columns" must be a list!'
                + "Is currently: type {0}".format(str(type(update_on_columns)))
            )
        field_constraints_mapping = EC.QBC_FIELD_CONSTRAINTS_MAPPING.get(
            self.dwh_engine,
        )

        # don't add primary key to create table definition
        # instead make alter table call later for the case of composite PKs
        field_constraints_mapping.pop(EC.QBC_FIELD_PK, None)
        pk_columns = []

        hash_columns = hash_columns or []
        for column_name in columns_definition.keys():
            raw_row[column_name] = None
            definition = columns_definition[column_name]
            if not isinstance(definition, dict):
                raise Exception(
                    "Column {0} is not properly defined!".format(
                        column_name,
                    )
                )
            sql_part_columns += [
                '"{0}"\t{1}\t{2}'.format(
                    column_name,  # Field name
                    self._get_column_type(definition),  # Type: text, int etc.
                    " ".join(
                        [  # Get all additional field properties (unique etc.)
                            field_constraints_mapping[addon]
                            if definition.get(addon)
                            else ""
                            for addon in list(field_constraints_mapping.keys())
                        ]
                    ),
                )
            ]
            if self._get_column_type(definition) == "jsonb":
                jsonb_columns += [column_name]
            if definition.get(EC.QBC_FIELD_PK):
                pk_columns += [column_name]
                if create_update_on_columns:
                    update_on_columns += [column_name]
            if definition.get(EC.QBC_FIELD_HASH):
                hash_columns += [column_name]
        sql_part_columns = ",\n\t".join(sql_part_columns)

        if clean_data_before_upload:
            self.log.info("Cleaning data for upload...")
            upload_data = []
            cols_list = list(raw_row.keys())
            if hash_columns:
                hash_func = getattr(hashlib, hashlib_func_name)
            while data:
                datum = data.pop(0)
                # Make sure that each dict in upload_data has all keys
                row = deepcopy(raw_row)
                for column_name, value in datum.items():
                    if column_name in cols_list:
                        if column_name in hash_columns:
                            # hash column!
                            pre_digest = hash_func(str(value).encode())
                            row[column_name] = pre_digest.hexdigest()
                        elif not value is None:
                            # avoid edge case of data where all instances of a
                            #   field are None, thus having data for a field
                            #   missing in the columns_definition!
                            if (
                                column_name in jsonb_columns
                                or isinstance(value, (dict, OrderedDict, list))
                                or (not isinstance(value, mapped_types))
                            ):
                                try:
                                    row[column_name] = json.dumps(
                                        value, cls=EWAHJSONEncoder
                                    )
                                except TypeError:
                                    # try dumping with bson utility function
                                    row[column_name] = dumps(value)
                            elif isinstance(value, str):
                                if not (value == "\0"):
                                    row[column_name] = value.replace("\x00", "")
                            else:
                                row[column_name] = value
                upload_data += [row]
            data_len = len(upload_data)
        else:
            data_len = len(data)
            upload_data = None

        self.log.info(
            "Uploading {0} rows of data...".format(
                str(data_len),
            )
        )

        kwargs = {}
        if database_name:
            kwargs["database_name"] = database_name

        kwargs.update(
            {
                "data": upload_data or data,
                "table_name": table_name,
                "schema_name": schema_name,
                "schema_suffix": schema_suffix,
                "columns_definition": columns_definition,
                "columns_partial_query": sql_part_columns,
                "update_on_columns": update_on_columns,
                "drop_and_replace": drop_and_replace,
                "pk_columns": pk_columns,
            }
        )
        self._create_or_update_table(**kwargs)

        if commit:
            self.dwh_hook.commit()



from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

from ewah.uploaders import get_uploader
from ewah.hooks.base import EWAHBaseHook
from ewah.constants import EWAHConstants as EC
from ewah.ewah_utils.ssh_tunnel import start_ssh_tunnel
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments as ada
from ewah.ewah_utils.airflow_utils import datetime_utcnow_with_tz

from datetime import datetime, timedelta
import time
import copy
import hashlib


class EWAHBaseOperator(BaseOperator):
    """Extension of airflow's native Base Operator.

    EWAH operators always work in the same way: They extract raw data from a
    source and load it into a relational database aka data warehouse aka DWH.
    Transformations only happen insofar as they are required to bring data
    into a relational format.

    *How to use*
    The child class extracts data from the source. At the end of the child
    class' execute() function, it calls self.update(). Optional: The child can
    also repeatedly call the self.update() class for chunking.

    *Arguments of self.update():*
    Data is a list of dictionaries, where each list element will become
    one row of data and the contained dict is fieldname:value. Value can be
    None, or a fieldname can be missing in a particular dictionary, in which
    case the value is assumed to be None.

    columns_definition may be supplied at the self.update() function call,
    at initialization, or not at all, depending on the usecase. The use order is
    - If available, take the columns_definition from the self.update() call
    - If None, check if columns_definition was supplied to the operator at init
    - If neither, create a columns_definition on the fly using (all) the data

    columns_definition is a dictionary with fieldname:properties. Properties
    is None or a dictionary of option:value where option can be one of the
    following:
    - EC.QBC_FIELD_TYPE -> String: Field type (default text)
    - EC.QBC_FIELD_PK -> Boolean: Is this field the primary key? (default False)

    Note that the value of EC.QBC_FIELD_TYPE is DWH-engine specific!

    columns_definition is case sensitive.

    Implemented DWH engines:
    - PostgreSQL
    - Snowflake

    Next up:
    - BigQuery
    - Redshift
    """

    # Overwrite _NAMES with a list of names that work as aliases when defining
    # DAGs via YAML
    _NAMES = []

    # in child classes, don't overwrite this but add values within __init__!
    template_fields = (
        "load_data_from",
        "load_data_until",
    )

    # Child class must update or overwrite these values
    # A missing element is interpreted as False
    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: False,
        EC.ES_INCREMENTAL: False,
        EC.ES_GET_NEXT: False,
    }

    _REQUIRES_COLUMNS_DEFINITION = False  # raise error if true and None supplied

    _CONN_TYPE = None  # overwrite me with the required connection type, if applicable

    _INDEX_QUERY = """
        CREATE INDEX IF NOT EXISTS {0}
        ON "{1}"."{2}" ({3})
    """

    upload_call_count = 0

    _metadata = {}  # to be updated by operator, if applicable

    def __init__(
        self,
        source_conn_id,
        dwh_engine,
        dwh_conn_id,
        extract_strategy,
        target_table_name,
        target_schema_name,
        target_schema_suffix="_next",
        target_database_name=None,  # Only for Snowflake
        load_data_from=None,  # set a minimum date e.g. for reloading of data
        reload_data_from=None,  # load data from this date for new tables
        load_data_from_relative=None,  # optional timedelta for incremental
        load_data_until=None,  # set a maximum date
        load_data_until_relative=None,  # optional timedelta for incremental
        load_data_chunking_timedelta=None,  # optional timedelta to chunk by
        columns_definition=None,
        update_on_columns=None,
        primary_key_column_name=None,
        clean_data_before_upload=True,
        exclude_columns=[],  # list of columns to exclude, if no
        # columns_definition was supplied (e.g. for select * with sql)
        index_columns=[],  # list of columns to create an index on. can be
        # an expression, must be quoted in list if quoting is required.
        source_ssh_tunnel_conn_id=None,  # create SSH tunnel if set; uses host
        # and port from source_conn_id as remote host and port
        target_ssh_tunnel_conn_id=None,  # see source_ssh_tunnel_conn_id
        hash_columns=None,  # str or list of str - columns to hash pre-upload
        hashlib_func_name="sha256",  # specify hashlib hashing function
        wait_for_seconds=120,  # seconds past next_execution_date to wait until
        # wait_for_seconds only applies for incremental loads
        add_metadata=True,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        for item in EWAHBaseOperator.template_fields:
            # Make sure template_fields was not overwritten
            _msg = "Operator must not overwrite template_fields!"
            assert item in self.template_fields, _msg

        _msg = 'param "wait_for_seconds" must be a nonnegative integer!'
        assert isinstance(wait_for_seconds, int) and wait_for_seconds >= 0, _msg
        _msg = "extract_strategy {0} not accepted for this operator!".format(
            extract_strategy,
        )
        assert self._ACCEPTED_EXTRACT_STRATEGIES.get(extract_strategy), _msg

        if hash_columns and not clean_data_before_upload:
            _msg = "column hashing is only possible with data cleaning!"
            raise Exception(_msg)
        elif isinstance(hash_columns, str):
            hash_columns = [hash_columns]
        if hashlib_func_name:
            _msg = "Invalid hashing function: hashlib.{0}()"
            _msg = _msg.format(hashlib_func_name)
            assert hasattr(hashlib, hashlib_func_name), _msg

        if columns_definition and exclude_columns:
            raise Exception(
                "Must not supply both columns_definition and " + "exclude_columns!"
            )

        if not dwh_engine or not dwh_engine in EC.DWH_ENGINES:
            _msg = "Invalid DWH Engine: {0}\n\nAccepted Engines:\n\t{1}".format(
                str(dwh_engine),
                "\n\t".join(EC.DWH_ENGINES),
            )
            raise Exception(_msg)

        if index_columns and not dwh_engine == EC.DWH_ENGINE_POSTGRES:
            raise Exception("Indices are only allowed for PostgreSQL DWHs!")

        if dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            if not target_database_name:
                conn_db_name = EWAHBaseHook.get_connection(dwh_conn_id)
                conn_db_name = conn_db_name.extra_dejson.get("database")
                if conn_db_name:
                    target_database_name = conn_db_name
                else:
                    raise Exception(
                        "If using DWH Engine {0}, must provide {1}!".format(
                            dwh_engine,
                            '"target_database_name" to specify the Database',
                        )
                    )
        else:
            if target_database_name:
                raise Exception('Received argument for "target_database_name"!')

        if self._REQUIRES_COLUMNS_DEFINITION:
            if not columns_definition:
                raise Exception(
                    "This operator requires the argument " + "columns_definition!"
                )

        if primary_key_column_name and update_on_columns:
            raise Exception(
                "Cannot supply BOTH primary_key_column_name AND" + " update_on_columns!"
            )

        if not extract_strategy in [EC.ES_FULL_REFRESH]:
            # Required settings for incremental loads
            # Update condition for new load strategies as required
            if not (
                update_on_columns
                or primary_key_column_name
                or (
                    columns_definition
                    and (
                        0
                        < sum(
                            [
                                bool(columns_definition[col].get(EC.QBC_FIELD_PK))
                                for col in list(columns_definition.keys())
                            ]
                        )
                    )
                )
            ):
                raise Exception(
                    "If this is incremental loading of a table, "
                    + "one of the following is required:"
                    + "\n- List of columns to update on (update_on_columns)"
                    + "\n- Name of the primary key (primary_key_column_name)"
                    + "\n- Column definition (columns_definition) that includes"
                    + " the primary key(s)"
                )

        _msg = "load_data_from_relative and load_data_until_relative must be"
        _msg += " timedelta if supplied!"
        assert isinstance(
            load_data_from_relative,
            (type(None), timedelta),
        ), _msg
        assert isinstance(
            load_data_until_relative,
            (type(None), timedelta),
        ), _msg
        _msg = "load_data_chunking_timedelta must be timedelta!"
        assert isinstance(
            load_data_chunking_timedelta,
            (type(None), timedelta),
        ), _msg

        self.source_conn_id = source_conn_id
        self.dwh_engine = dwh_engine
        self.dwh_conn_id = dwh_conn_id
        self.extract_strategy = extract_strategy
        self.target_table_name = target_table_name
        self.target_schema_name = target_schema_name
        self.target_schema_suffix = target_schema_suffix
        self.target_database_name = target_database_name
        self.load_data_from = load_data_from
        self.reload_data_from = reload_data_from
        self.load_data_from_relative = load_data_from_relative
        self.load_data_until = load_data_until
        self.load_data_until_relative = load_data_until_relative
        self.load_data_chunking_timedelta = load_data_chunking_timedelta
        self.columns_definition = columns_definition
        if (not update_on_columns) and primary_key_column_name:
            if type(primary_key_column_name) == str:
                update_on_columns = [primary_key_column_name]
            elif type(primary_key_column_name) in (list, tuple):
                update_on_columns = primary_key_column_name
        self.update_on_columns = update_on_columns
        self.clean_data_before_upload = clean_data_before_upload
        self.primary_key_column_name = primary_key_column_name  # may be used ...
        #   ... by a child class at execution!
        self.exclude_columns = exclude_columns
        self.index_columns = index_columns
        self.source_ssh_tunnel_conn_id = source_ssh_tunnel_conn_id
        self.target_ssh_tunnel_conn_id = target_ssh_tunnel_conn_id
        self.hash_columns = hash_columns
        self.hashlib_func_name = hashlib_func_name
        self.wait_for_seconds = wait_for_seconds
        self.add_metadata = add_metadata

        self.uploader = get_uploader(self.dwh_engine)

        _msg = "DWH hook does not support extract strategy {0}!".format(
            extract_strategy,
        )
        # assert self.uploader._ACCEPTED_EXTRACT_STRATEGIES.get(extract_strategy), _msg

    def ewah_execute(self, context):
        raise Exception("You need to overwrite me!")

    def execute(self, context):
        """Why this method is defined here:
        When executing a task, airflow calls this method. Generally, this
        method contains the "business logic" of the individual operator.
        However, EWAH may want to do some actions for all operators. Thus,
        the child operators shall have an ewah_execute() function which is
        called by this general execute() method.
        """

        def close_ssh_tunnels():
            # close SSH tunnels if they exist
            if hasattr(self, "source_ssh_tunnel_forwarder"):
                self.source_ssh_tunnel_forwarder.stop()
                del self.source_ssh_tunnel_forwarder
            if hasattr(self, "target_ssh_tunnel_forwarder"):
                self.target_ssh_tunnel_forwarder.stop()
                del self.target_ssh_tunnel_forwarder

        # required for metadata in data upload
        self._execution_time = datetime_utcnow_with_tz()
        self._context = context

        # the upload hook is used in the self.upload_data() function
        # which is called by the child's ewah_execute function whenever there is
        # data to upload. If applicable: start SSH tunnel first!
        if self.target_ssh_tunnel_conn_id:
            self.log.info("Opening SSH tunnel to target...")
            self.target_ssh_tunnel_forwarder, self.dwh_conn = start_ssh_tunnel(
                ssh_conn_id=self.target_ssh_tunnel_conn_id,
                remote_conn_id=self.dwh_conn_id,
            )
        else:
            self.dwh_conn = EWAHBaseHook.get_connection(self.dwh_conn_id)
        self.uploader = self.uploader(self.dwh_conn)

        # open SSH tunnel for the data source connection, if applicable
        if self.source_ssh_tunnel_conn_id:
            self.log.info("Opening SSH tunnel to source...")
            self.source_ssh_tunnel_forwarder, self.source_conn = start_ssh_tunnel(
                ssh_conn_id=self.source_ssh_tunnel_conn_id,
                remote_conn_id=self.source_conn_id,
            )
        elif self.source_conn_id:
            # resolve conn id here & delete the object to avoid usage elsewhere
            self.source_conn = EWAHBaseHook.get_connection(self.source_conn_id)
        if hasattr(self, "source_conn"):
            self.source_hook = self.source_conn.get_hook()
        del self.source_conn_id

        if self._CONN_TYPE:
            _msg = "Error - connection type must be {0}!".format(self._CONN_TYPE)
            assert self._CONN_TYPE == self.source_conn.conn_type, _msg

        temp_schema_name = self.target_schema_name + self.target_schema_suffix
        # Create a new copy of the target table.
        # This is so data is loaded into a new table and if data loading
        # fails, the original data is not corrupted. At a new try or re-run,
        # the original table is just copited anew.
        if not self.extract_strategy == EC.ES_FULL_REFRESH:
            # Full refresh always drops and replaces the tables completely
            self.uploader.copy_table(
                old_schema=self.target_schema_name,
                old_table=self.target_table_name,
                new_schema=temp_schema_name,
                new_table=self.target_table_name,
                database_name=self.target_database_name,
            )

        # set load_data_from and load_data_until as required
        data_from = ada(self.load_data_from)
        data_until = ada(self.load_data_until)
        if self.extract_strategy == EC.ES_INCREMENTAL:
            _tdz = timedelta(days=0)  # aka timedelta zero
            _ed = context["execution_date"]
            _ned = context["next_execution_date"]

            # normal incremental load
            _ed -= self.load_data_from_relative or _tdz
            data_from = max(_ed, data_from or _ed)
            if not self.test_if_target_table_exists():
                # Load data from scratch!
                data_from = ada(self.reload_data_from) or data_from

            _ned += self.load_data_until_relative or _tdz
            data_until = min(_ned, data_until or _ned)

        elif self.extract_strategy == EC.ES_FULL_REFRESH:
            # Values may still be set as static values
            data_from = ada(self.reload_data_from) or data_from

        else:
            _msg = "Must define load_data_from etc. behavior for load strategy!"
            raise Exception(_msg)

        self.data_from = data_from
        self.data_until = data_until
        # del variables to make sure they are not used later on
        del self.load_data_from
        del self.reload_data_from
        del self.load_data_until
        del self.load_data_from_relative
        del self.load_data_until_relative

        # Have an option to wait until a short period (e.g. 2 minutes) past
        # the incremental loading range timeframe to ensure that all data is
        # loaded, useful e.g. if APIs lag or if server timestamps are not
        # perfectly accurate.
        # When a DAG is executed as soon as possible, some data sources
        # may not immediately have up to date data from their API.
        # E.g. querying all data until 12.30pm only gives all relevant data
        # after 12.32pm due to some internal delays. In those cases, make
        # sure the (incremental loading) DAGs don't execute too quickly.
        if self.wait_for_seconds and self.extract_strategy == EC.ES_INCREMENTAL:
            wait_until = context.get("next_execution_date")
            if wait_until:
                wait_until += timedelta(seconds=self.wait_for_seconds)
                self.log.info(
                    "Awaiting execution until {0}...".format(
                        str(wait_until),
                    )
                )
            while wait_until and datetime_utcnow_with_tz() < wait_until:
                # Only sleep a maximum of 5s at a time
                wait_for_timedelta = wait_until - datetime_utcnow_with_tz()
                time.sleep(max(0, min(wait_for_timedelta.total_seconds(), 5)))

        try:
            # execute operator
            if self.load_data_chunking_timedelta and data_from and data_until:
                # Chunking to avoid OOM
                assert data_until > data_from
                assert self.load_data_chunking_timedelta > timedelta(days=0)
                while self.data_from < data_until:
                    self.data_until = self.data_from
                    self.data_until += self.load_data_chunking_timedelta
                    self.data_until = min(self.data_until, data_until)
                    self.ewah_execute(context)
                    self.data_from += self.load_data_chunking_timedelta
            else:
                self.ewah_execute(context)

            # if PostgreSQL and arg given: create indices
            for column in self.index_columns:
                assert self.dwh_engine == EC.DWH_ENGINE_POSTGRES
                # Use hashlib to create a unique 63 character string as index
                # name to avoid breaching index name length limits & accidental
                # duplicates / missing indices due to name truncation leading to
                # identical index names.
                self.hook.execute(
                    self._INDEX_QUERY.format(
                        "__ewah_"
                        + hashlib.blake2b(
                            (
                                temp_schema_name
                                + "."
                                + self.target_table_name
                                + "."
                                + column
                            ).encode(),
                            digest_size=28,
                        ).hexdigest(),
                        self.target_schema_name + self.target_schema_suffix,
                        self.target_table_name,
                        column,
                    )
                )

            # commit only at the end, so that no data may be committed before an
            # error occurs.
            self.log.info("Now committing changes!")
            self.uploader.commit()
            self.uploader.close()
        except:
            # close SSH tunnels on failure before raising the error
            close_ssh_tunnels()
            raise

        # everything worked, now close SSH tunnels
        close_ssh_tunnels()

    def test_if_target_table_exists(self):
        # Need to use existing hook to work within open transaction
        kwargs = {
            "table_name": self.target_table_name,
            "schema_name": self.target_schema_name + self.target_schema_suffix,
        }
        if self.dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            kwargs["database_name"] = self.target_database_name
            return self.uploader.test_if_table_exists(**kwargs)
        if self.dwh_engine in [EC.DWH_ENGINE_POSTGRES, EC.DWH_ENGINE_SNOWFLAKE]:
            return self.uploader.test_if_table_exists(**kwargs)
        # For a new DWH, need to manually check if function works properly
        # Thus, fail until explicitly added
        raise Exception("Function not implemented!")

    def _create_columns_definition(self, data):
        "Create a columns_definition from data (list of dicts)."
        inconsistent_data_type = EC.QBC_TYPE_MAPPING[self.dwh_engine].get(
            EC.QBC_TYPE_MAPPING_INCONSISTENT
        )

        def get_field_type(value):
            return (
                EC.QBC_TYPE_MAPPING[self.dwh_engine].get(type(value))
                or inconsistent_data_type
            )

        result = {}
        for datum in data:
            for field in datum.keys():
                if field in self.exclude_columns:
                    datum[field] = None
                elif field in (self.hash_columns or []) and not result.get(field):
                    # Type is appropriate string type & QBC_FIELD_HASH is true
                    result.update(
                        {
                            field: {
                                EC.QBC_FIELD_TYPE: get_field_type("str"),
                                EC.QBC_FIELD_HASH: True,
                            }
                        }
                    )
                elif not (
                    result.get(field, {}).get(EC.QBC_FIELD_TYPE)
                    == inconsistent_data_type
                ) and (not datum[field] is None):
                    if result.get(field):
                        # column has been added in a previous iteration.
                        # If not default column: check if new and old column
                        #   type identification agree.
                        if not (
                            result[field][EC.QBC_FIELD_TYPE]
                            == get_field_type(datum[field])
                        ):
                            self.log.info(
                                "WARNING! Data types are inconsistent."
                                + " Affected column: {0}".format(field)
                            )
                            result[field][EC.QBC_FIELD_TYPE] = inconsistent_data_type

                    else:
                        # First iteration with this column. Add to result.
                        result.update(
                            {field: {EC.QBC_FIELD_TYPE: get_field_type(datum[field])}}
                        )
        return result

    def upload_data(self, data=None, columns_definition=None):
        """Upload data, no matter the source. Call this functions in the child
        operator whenever data is available for upload, as often as needed.
        """
        if not data:
            self.log.info("No data to upload!")
            return
        self.upload_call_count += 1
        self.log.info(
            "Chunk {1}: Uploading {0} rows of data.".format(
                str(len(data)),
                str(self.upload_call_count),
            )
        )

        if self.add_metadata:
            self.log.info("Adding metadata...")
            metadata = copy.deepcopy(self._metadata)  # from individual operator
            # for all operators alike
            metadata.update(
                {
                    "_ewah_executed_at": self._execution_time,
                    "_ewah_execution_chunk": self.upload_call_count,
                    "_ewah_dag_id": self._context["dag"].dag_id,
                    "_ewah_dag_run_id": self._context["run_id"],
                    "_ewah_dag_run_execution_date": self._context["execution_date"],
                    "_ewah_dag_run_next_execution_date": self._context[
                        "next_execution_date"
                    ],
                }
            )
            for datum in data:
                datum.update(metadata)

        columns_definition = columns_definition or self.columns_definition
        if not columns_definition:
            self.log.info("Creating table schema on the fly based on data.")
            columns_definition = self._create_columns_definition(data)

        if self.update_on_columns:
            pk_list = self.update_on_columns  # is a list already
        elif self.primary_key_column_name:
            pk_list = [self.primary_key_column_name]
        else:
            pk_list = []

        if pk_list:
            for pk_name in pk_list:
                if not pk_name in columns_definition.keys():
                    raise Exception(
                        ("Column {0} does not exist but is " + "expected!").format(
                            pk_name
                        )
                    )
                columns_definition[pk_name][EC.QBC_FIELD_PK] = True

        hook = self.uploader

        if (self.extract_strategy == EC.ES_INCREMENTAL) or (self.upload_call_count > 1):
            self.log.info("Checking for, and applying schema changes.")
            _new_schema_name = self.target_schema_name + self.target_schema_suffix
            new_cols, del_cols = hook.detect_and_apply_schema_changes(
                new_schema_name=_new_schema_name,
                new_table_name=self.target_table_name,
                new_columns_dictionary=columns_definition,
                # When introducing a feature utilizing this, remember to
                #  consider multiple runs within the same execution
                drop_missing_columns=False and self.upload_call_count == 1,
                database=self.target_database_name,
                commit=False,  # Commit only when / after uploading data
            )
            self.log.info(
                "Added fields:\n\t{0}\nDeleted fields:\n\t{1}".format(
                    "\n\t".join(new_cols) or "\n",
                    "\n\t".join(del_cols) or "\n",
                )
            )

        self.log.info("Uploading data now.")
        hook.create_or_update_table(
            data=data,
            columns_definition=columns_definition,
            table_name=self.target_table_name,
            schema_name=self.target_schema_name,
            schema_suffix=self.target_schema_suffix,
            database_name=self.target_database_name,
            drop_and_replace=(self.extract_strategy == EC.ES_FULL_REFRESH)
            and (self.upload_call_count == 1),  # In case of chunking of uploads
            update_on_columns=self.update_on_columns,
            commit=False,  # See note below for reason
            clean_data_before_upload=self.clean_data_before_upload,
            hash_columns=self.hash_columns,
            hashlib_func_name=self.hashlib_func_name,
        )
        """ Note on committing changes:
            The hook used for data uploading is created at the beginning of the
            execute function and automatically committed and closed at the end.
            DO NOT commit in this function, as multiple uploads may be required,
            and any intermediate commit may be subsequently followed by an
            error, which would then result in incomplete data committed.
        """


class EWAHEmptyOperator(EWAHBaseOperator):
    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
    }

    def __init__(self, *args, **kwargs):
        raise Exception(
            "Failed to load operator! Probably missing"
            + " requirements for the operator in question.\n\nSupplied args:"
            + "\n\t"
            + "\n\t".join(args)
            + "\n\nSupplied kwargs:\n\t"
            + "\n\t".join(["{0}: {1}".format(k, v) for k, v in kwargs.items()])
        )



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from airflow.providers.amazon.aws.hooks.s3 import S3Hook

from datetime import datetime, timedelta

import json
import csv
import gzip


class EWAHS3Operator(EWAHBaseOperator):
    """Only implemented for JSON and CSV files from S3 right now!"""

    _NAMES = ["s3"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
    }

    _IMPLEMENTED_FORMATS = [
        "JSON",
        "AWS_FIREHOSE_JSON",
        "CSV",
    ]

    _BOM = b"\xef\xbb\xbf"

    def __init__(
        self,
        bucket_name,
        file_format,
        prefix="",  # use for subfolder structures
        suffix=None,  # use e.g. for file types
        key_name=None,  # if not specified, uses data_from and data_until
        csv_format_options={},
        csv_encoding="utf-8",
        decompress=False,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        if not file_format in self._IMPLEMENTED_FORMATS:
            raise Exception(
                "File format not implemented! Available formats:"
                + " {0}".format(str(self._IMPLEMENTED_FORMATS))
            )

        if key_name:
            if self.load_data_from or self.load_data_until:
                raise Exception(
                    "Both key_name and either data_from or data_"
                    + "until specified! Cannot have both specified."
                )

        if not file_format == "CSV" and csv_format_options:
            raise Exception("csv_format_options is only valid for CSV files!")

        if decompress and not file_format == "CSV":
            raise exception("Can currently only decompress CSVs!")

        self.bucket_name = bucket_name
        self.prefix = prefix
        self.suffix = suffix
        self.key_name = key_name
        self.file_format = file_format
        self.csv_format_options = csv_format_options
        self.csv_encoding = csv_encoding
        self.decompress = decompress

    def _iterate_through_bucket(self, s3hook, bucket, prefix):
        """The bucket.objects.filter() method only returns a max of 1000
        objects. If more objects are in an S3 bucket, pagniation is
        required. See also: https://stackoverflow.com/questions/44238525/how-to-iterate-over-files-in-an-s3-bucket
        """
        cli = s3hook.get_client_type("s3")
        paginator = cli.get_paginator("list_objects_v2")
        page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)

        for page in page_iterator:
            if page["KeyCount"] > 0:
                for item in page["Contents"]:
                    yield item

    def ewah_execute(self, context):
        if self.file_format == "JSON":
            return self.execute_json(
                context=context,
                f_get_data=lambda hook, key, bucket: json.loads(
                    hook.read_key(key, bucket)
                ),
            )
        elif self.file_format == "AWS_FIREHOSE_JSON":
            return self.execute_json(
                context=context,
                f_get_data=lambda hook, key, bucket: json.loads(
                    "[" + hook.read_key(key, bucket).replace("}{", "},{") + "]"
                ),
            )
        elif self.file_format == "CSV":
            return self.execute_csv(
                context=context,
            )
        else:
            raise Exception("File format not implemented!")

    def execute_csv(self, context):
        hook = S3Hook(self.source_conn.conn_id)
        suffix = self.suffix
        if self.key_name:
            obj = hook.get_key(self.key_name, self.bucket_name)
            raw_data = obj.get()["Body"].read()
            if self.decompress:
                raw_data = gzip.decompress(raw_data)
            if raw_data[:3] == self._BOM:
                raw_data = raw_data[3:]
                csv_encoding = "utf-8"
            else:
                csv_encoding = self.csv_encoding
            try:
                raw_data = raw_data.decode(csv_encoding).splitlines()
            except UnicodeDecodeError:
                if csv_encoding == self.csv_encoding:
                    raise
                raw_data = raw_data.decode(self.csv_encoding).splitlines()
            reader = csv.DictReader(raw_data, **self.csv_format_options)
            data = list(reader)
            self.upload_data(data=data)
        else:
            objects = self._iterate_through_bucket(
                s3hook=hook,
                bucket=self.bucket_name,
                prefix=self.prefix,
            )
            for obj_iter in objects:
                # skip all files outside of loading scope
                obj = hook.get_key(obj_iter["Key"], self.bucket_name)
                if self.data_from and (obj.last_modified < self.data_from):
                    continue
                if self.data_until and (obj.last_modified >= self.data_until):
                    continue
                if suffix and not suffix == obj.key[-len(suffix) :]:
                    continue

                self.log.info("Loading data from file {0}".format(obj.key))
                raw_data = obj.get()["Body"].read()
                if self.decompress:
                    raw_data = gzip.decompress(raw_data)
                # remove BOM if it exists
                # also, if file has a BOM, it is 99.9% utf-8 encoded!
                if raw_data[:3] == self._BOM:
                    raw_data = raw_data[3:]
                    csv_encoding = "utf-8"
                else:
                    csv_encoding = self.csv_encoding
                # there may be a BOM while still not utf-8 -> use the
                # given csv_encoding argument if so
                try:
                    raw_data = raw_data.decode(csv_encoding).splitlines()
                except UnicodeDecodeError:
                    if csv_encoding == self.csv_encoding:
                        raise
                    raw_data = raw_data.decode(self.csv_encoding).splitlines()
                reader = csv.DictReader(raw_data, **self.csv_format_options)
                data = list(reader)
                self._metadata.update(
                    {
                        "bucket_name": self.bucket_name,
                        "file_name": obj.key,
                        "file_last_modified": str(obj.last_modified),
                    }
                )
                self.upload_data(data=data)

    def execute_json(self, context, f_get_data):
        hook = S3Hook(self.source_conn.conn_id)
        suffix = self.suffix

        if self.key_name:
            data = f_get_data(
                hook=hook,
                key=self.key_name,
                bucket=self.bucket_name,
            )
            self.upload_data(data=data)
        else:
            data = []
            objects = self._iterate_through_bucket(
                s3hook=hook,
                bucket=self.bucket_name,
                prefix=self.prefix,
            )
            for obj_iter in objects:
                obj = hook.get_key(obj_iter["Key"], self.bucket_name)
                if self.load_data_from and obj.last_modified < self.load_data_from:
                    continue
                if self.load_data_until and obj.last_modified >= self.load_data_until:
                    continue
                if suffix and not suffix == obj.key[-len(suffix) :]:
                    continue

                self.log.info(
                    "Loading data from file {0}".format(
                        obj.key,
                    )
                )
                self._metadata.update(
                    {
                        "bucket_name": self.bucket_name,
                        "file_name": obj.key,
                        "file_last_modified": str(obj.last_modified),
                    }
                )
                data = f_get_data(
                    hook=hook,
                    key=obj.key,
                    bucket=self.bucket_name,
                )
                self.upload_data(data=data)



from ewah.operators.sql_base import EWAHSQLBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

import pymysql


class EWAHMySQLOperator(EWAHSQLBaseOperator):

    _NAMES = ["mysql"]

    _SQL_BASE = "SELECT\n{columns}\nFROM `{schema}`.`{table}`\nWHERE {where_clause}"
    _SQL_BASE_SELECT = "SELECT * FROM ({select_sql}) t WHERE {{0}}"
    _SQL_COLUMN_QUOTE = "`"
    _SQL_MINMAX_CHUNKS = """
        SELECT MIN({column}), MAX({column}) FROM ({base}) t
    """
    _SQL_CHUNKING_CLAUSE = """
        AND {column} >= %(from_value)s
        AND {column} <{equal_sign} %(until_value)s
    """
    _SQL_PARAMS = "%({0})s"

    def _get_data_from_sql(self, sql, params=None, return_dict=True):
        if return_dict:
            cursor_class = pymysql.cursors.DictCursor
        else:
            cursor_class = pymysql.cursors.Cursor
        database_conn = pymysql.connect(
            **{
                "host": self.source_conn.host,
                "user": self.source_conn.login,
                "passwd": self.source_conn.password,
                "port": self.source_conn.port,
                "database": self.source_conn.schema,
                "cursorclass": cursor_class,
            }
        )
        cursor = database_conn.cursor()
        self.log.info(
            "Executing:\n{0}\n\nWith params:\n{1}".format(
                sql,
                str(params),
            )
        )
        cursor.execute(sql, args=params)
        data = cursor.fetchall()
        cursor.close()
        database_conn.close()
        return data



import os, sys
from ewah.hooks.base import EWAHBaseHook

# import all hooks by walking through all files in this directory and
# importing all objects that are subclasses of EWAHBaseHook
# adapted from:
#   https://stackoverflow.com/questions/6246458/import-all-classes-in-directory
relevant_files = [
    f[:-3]
    for f in os.listdir(os.path.dirname(os.path.abspath(__file__)))
    if f.endswith(".py") and not f in ["__init__.py"]
]

hook_class_names = []
name_template = "ewah.hooks.{0}.{1}"

for py_file in relevant_files:
    mod = __import__(".".join([__name__, py_file]), fromlist=[py_file])
    classes = [getattr(mod, x) for x in dir(mod) if isinstance(getattr(mod, x), type)]
    for cls in classes:
        if issubclass(cls, EWAHBaseHook) and not cls == EWAHBaseHook:
            hook_class_names.append(name_template.format(py_file, cls.__name__))

## Sample hook - delete this when all features have been tested

# from typing import Any, Dict, Optional
# from airflow.hooks.dbapi import DbApiHook
# from airflow.models.connection import Connection
#
#
# class JdbcHook(DbApiHook):
#     """
#     General hook for jdbc db access.
#     JDBC URL, username and password will be taken from the predefined connection.
#     Note that the whole JDBC URL must be specified in the "host" field in the DB.
#     Raises an airflow error if the given connection id doesn't exist.
#     """
#
#     conn_name_attr = 'jdbc_conn_id'
#     default_conn_name = 'jdbc_default'
#     conn_type = 'jdbc'
#     hook_name = 'JDBC Connection'
#     supports_autocommit = True
#
#     @staticmethod
#     def get_connection_form_widgets() -> Dict[str, Any]:
#         """Returns connection widgets to add to connection form"""
#         from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
#         from flask_babel import lazy_gettext
#         from wtforms import StringField
#
#         return {
#             "extra__jdbc__drv_path": StringField(lazy_gettext('Driver Path'), widget=BS3TextFieldWidget()),
#             "extra__jdbc__drv_clsname": StringField(
#                 lazy_gettext('Driver Class'), widget=BS3TextFieldWidget()
#             ),
#         }
#
#     @staticmethod
#     def get_ui_field_behaviour() -> Dict:
#         """Returns custom field behaviour"""
#         return {
#             "hidden_fields": ['port', 'schema', 'extra'],
#             "relabeling": {'host': 'Connection URL'},
#         }
#
#     def get_conn(self) -> Dict:
#         conn: Connection = self.get_connection(getattr(self, self.conn_name_attr))
#         host: str = conn.host
#         login: str = conn.login
#         psw: str = conn.password
#         jdbc_driver_loc: Optional[str] = conn.extra_dejson.get('extra__jdbc__drv_path')
#         jdbc_driver_name: Optional[str] = conn.extra_dejson.get('extra__jdbc__drv_clsname')
#
#         conn = {
#             "jclassname": jdbc_driver_name,
#             "url": str(host),
#             "driver_args": [str(login), str(psw)],
#             "jars": jdbc_driver_loc.split(",") if jdbc_driver_loc else None,
#         }
#         return conn



from ewah.hooks import hook_class_names
from ewah import VERSION


def get_provider_info():
    return {
        "package-name": "ewah",
        "name": "ewah",
        "description": "",
        "hook-class-names": hook_class_names,
        "versions": [VERSION],
    }



from ewah.operators.sql_base import EWAHSQLBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

from google.cloud import bigquery
from tempfile import NamedTemporaryFile
from datetime import timedelta

import os


class EWAHBigQueryOperator(EWAHSQLBaseOperator):

    _NAMES = ["bq", "biqguery"]

    _SQL_BASE = """
        SELECT
            {columns}
        FROM `{database}`.`{schema}`.`{table}`
        WHERE {where_clause}
    """
    _SQL_BASE_SELECT = "SELECT * FROM ({select_sql}) t WHERE {{0}}"
    _SQL_COLUMN_QUOTE = "`"
    _SQL_MINMAX_CHUNKS = """
        SELECT MIN({column}), MAX({column}) FROM ({base}) t
    """
    _SQL_CHUNKING_CLAUSE = """
        AND {column} >= @from_value
        AND {column} <{equal_sign} @until_value
    """
    _SQL_PARAMS = "@{0}"

    def __init__(self, *args, is_sharded=False, **kwargs):

        # Special case: getting a sharded table; then source_table_name shall be
        # the base name, e.g. events_ if getting events_*
        # To be refactored in the future!
        if is_sharded:
            assert False, "Sharding Feature not yet implemented!"
            if kwargs.get("extract_strategy") == EC.ES_FULL_REFRESH:
                self.is_sharded = False
                kwargs["source_table_name"] += "*"  # get all at full refresh!
            else:
                # These kwargs are not saved by default but are needed
                # if getting data from a sharded table
                self.is_sharded = True
                self.bq_table_name = kwargs["source_table_name"]
                self.bq_schema_name = kwargs["source_schema_name"]
                self.bq_dataset_name = kwargs["source_database_name"]
        else:
            self.is_sharded = False

        _msg = "Must supply source_database_name (=dataset name)!"
        assert kwargs.get("source_database_name"), _msg
        super().__init__(*args, **kwargs)

    def _get_data_from_sql(self, sql, params=None, return_dict=True):

        # set params
        if params:
            type_mapping = EC.QBC_TYPE_MAPPING.get(EC.DWH_ENGINE_BIGQUERY)
            _default = type_mapping[EC.QBC_TYPE_MAPPING_DEFAULT]
            job_config = bigquery.QueryJobConfig(
                query_parameters=[
                    bigquery.ScalarQueryParameter(
                        key,
                        type_mapping.get(type(value), _default),
                        value,
                    )
                    for key, value in params.items()
                ]
            )
        else:
            job_config = None

        # create temp file with creds
        self.log.info(
            "Executing:\n{0}\n\nWith params:\n{1}".format(
                sql,
                str(params),
            )
        )
        with NamedTemporaryFile() as cred_json:
            cred_filename = os.path.abspath(cred_json.name)
            cred_json.seek(0)
            cred_json.write(self.source_conn.extra.encode())
            cred_json.seek(0)
            cred_json.flush()
            client = bigquery.Client.from_service_account_json(cred_filename)
            results = client.query(sql, job_config=job_config).result()
            dict_result = [dict(row) for row in results]

        if return_dict or not dict_result:
            return dict_result
        keys = dict_result[0].keys()
        return [[row[key] for key in keys] for row in dict_result]

    def ewah_execute(self, context):
        # If getting a sharded events table, use a different approach

        if not self.is_sharded:
            # This should also run in the case of the full refresh DAG of a
            # fullcremental strategy even if it is a sharded table
            return super().ewah_execute(context)

        # code below only runs if is_sharded
        # loop through the relevant days day by day
        date_from = self.load_data_from.date()
        date_until = self.load_data_until.date()
        while date_from <= date_until:
            self.upload_data(
                data=self._get_data_from_sql(
                    sql=self._SQL_BASE.format(
                        columns="*",
                        database=self.bq_dataset_name,
                        schema=self.bq_schema_name,
                        table=self.bq_table_name + date_from.strftime("%Y%m%d"),
                        where_clause="1=1",
                    ),
                    return_dict=True,
                )
            )

            date_from += timedelta(days=1)



from airflow import DAG

from ewah.constants import EWAHConstants as EC
from ewah.ewah_utils.airflow_utils import etl_schema_tasks
from ewah.operators.base import EWAHBaseOperator

from collections.abc import Iterable
from copy import deepcopy
from datetime import datetime, timedelta
from typing import Optional, Type, Callable, List, Tuple, Union

import re


def dag_factory_drop_and_replace(
    dag_name: str,
    dwh_engine: str,
    dwh_conn_id: str,
    start_date: datetime,
    el_operator: Type[EWAHBaseOperator],
    operator_config: dict,
    target_schema_name: str,
    target_schema_suffix: str = "_next",
    target_database_name: Optional[str] = None,
    default_args: Optional[dict] = None,
    schedule_interval: timedelta = timedelta(days=1),
    end_date: Optional[datetime] = None,
    read_right_users: Optional[Union[List[str], str]] = None,
    dwh_ssh_tunnel_conn_id: Optional[str] = None,
    additional_dag_args: Optional[dict] = None,
    additional_task_args: Optional[dict] = None,
    logging_func: Optional[Callable] = None,
    **kwargs
) -> Tuple[DAG]:
    def raise_exception(msg: str) -> None:
        """Add information to error message before raising."""
        raise Exception("DAG: {0} - Error: {1}".format(dag_name, msg))

    logging_func = logging_func or print

    if kwargs:
        logging_func("unused config: {0}".format(str(kwargs)))

    additional_dag_args = additional_dag_args or {}
    additional_task_args = additional_task_args or {}

    if dwh_ssh_tunnel_conn_id and not dwh_engine == EC.DWH_ENGINE_POSTGRES:
        raise_exception("DWH tunneling only implemented for PostgreSQL DWHs!")
    if not read_right_users is None:
        if isinstance(read_right_users, str):
            read_right_users = [u.strip() for u in read_right_users.split(",")]
        if not isinstance(read_right_users, Iterable):
            raise_exception("read_right_users must be an iterable or string!")

    dag = DAG(
        dag_name,
        catchup=False,
        default_args=default_args,
        max_active_runs=1,
        schedule_interval=schedule_interval,
        start_date=start_date,
        end_date=end_date,
        **additional_dag_args,
    )

    kickoff, final = etl_schema_tasks(
        dag=dag,
        dwh_engine=dwh_engine,
        dwh_conn_id=dwh_conn_id,
        target_schema_name=target_schema_name,
        target_schema_suffix=target_schema_suffix,
        target_database_name=target_database_name,
        read_right_users=read_right_users,
        ssh_tunnel_conn_id=dwh_ssh_tunnel_conn_id,
        **additional_task_args,
    )

    base_config = deepcopy(additional_task_args)
    base_config.update(operator_config.get("general_config", {}))
    with dag:
        for table in operator_config["tables"].keys():
            table_config = deepcopy(base_config)
            table_config.update(operator_config["tables"][table] or {})
            table_config.update(
                {
                    "task_id": "extract_load_" + re.sub(r"[^a-zA-Z0-9_]", "", table),
                    "dwh_engine": dwh_engine,
                    "dwh_conn_id": dwh_conn_id,
                    "extract_strategy": EC.ES_FULL_REFRESH,
                    "target_table_name": operator_config["tables"][table].get(
                        "target_table_name", table
                    ),
                    "target_schema_name": target_schema_name,
                    "target_schema_suffix": target_schema_suffix,
                    "target_database_name": target_database_name,
                    "target_ssh_tunnel_conn_id": dwh_ssh_tunnel_conn_id,
                }
            )
            table_task = el_operator(**table_config)
            kickoff >> table_task >> final

    return (dag,)



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

import mailchimp3


class EWAHMailchimpOperator(EWAHBaseOperator):

    _NAMES = ["mailchimp", "mc"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
    }

    _REQUIRES_COLUMNS_DEFINITION = False

    def __init__(self, resource=None, *args, **kwargs):
        # API docs: https://mailchimp.com/developer/api/marketing/

        # use target table name as resource if none is given (-> easier config)
        resource = resource or kwargs.get("target_table_name")

        # validate resource name
        if not hasattr(mailchimp3.entities, resource):
            # see here for acceptable resources: https://github.com/VingtCinq/python-mailchimp/blob/master/mailchimp3/__init__.py
            raise Exception("Invalid resource {0}!".format(resource))

        self.resource = resource
        super().__init__(*args, **kwargs)

    def ewah_execute(self, context):
        # Initiate Mailchimp client
        self.log.info("Connecting to Mailchimp...")
        conn = self.source_conn
        mailchimp_cli = mailchimp3.MailChimp(
            mc_user=conn.login,
            mc_api=conn.password,
        )

        # get data!
        self.log.info("Loading data from Mailchimp...")
        resource = self.resource
        data = getattr(mailchimp_cli, resource).all(get_all=True)[resource]

        # upload data
        self.upload_data(data)



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

from sshtunnel import SSHTunnelForwarder
from tempfile import NamedTemporaryFile
from datetime import timedelta
from pytz import timezone

import os


class EWAHSQLBaseOperator(EWAHBaseOperator):

    # Don't overwrite the _NAMES list to avoid accidentally exposing the SQL
    # base operator
    # _NAMES = []

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
    }

    def __init__(
        self,
        source_schema_name=None,  # string
        source_table_name=None,  # string, defaults to same as target_table_name
        source_database_name=None,  # bigquery: project id
        sql_select_statement=None,  # SQL as alternative to source_table_name
        use_execution_date_for_incremental_loading=False,  # use context instead
        #   of data_from and data_until
        timestamp_column=None,  # name of column to increment and/or chunk by
        chunking_interval=None,  # can be datetime.timedelta or integer
        chunking_column=None,  # defaults to primary key if integer
        # also potentially used: primary_key_column_name of parent operator
        where_clause=None,  # restrict data loaded by operator
        *args,
        **kwargs
    ):
        where_clause = where_clause or "1 = 1"

        # allow setting schema in general config w/out throwing an error
        if sql_select_statement:
            source_schema_name = None
            source_database_name = None

        target_table_name = kwargs.get("target_table_name")

        if chunking_interval:
            if type(chunking_interval) == timedelta:
                if not timestamp_column:
                    raise Exception(
                        "If chunking via timedelta, must supply"
                        + " a timestamp_column, even if not loading "
                        + "incrementally!"
                    )
            elif type(chunking_interval) == int:
                if not chunking_column and not kwargs.get("primary_key_column_name"):
                    # Check columns for primary key - if exactly one,
                    # use it. Otherwise, raise error.
                    error_msg = (
                        "If chunking via integer, must supply "
                        + "primary_key_column_name OR have EXACTLY ONE "
                        + "primary key defined in the columns_definition dict"
                        + "! This is not the case for: {0}. {{0}}".format(
                            source_table_name
                        )
                    )
                    if not kwargs.get("columns_definition"):
                        raise Exception(
                            error_msg.format(
                                "You did not supply the columns_definition dict."
                            )
                        )
                    columns = kwargs.get("columns_definition")
                    if not (
                        sum(
                            [
                                1 if columns[key].get(EC.QBC_FIELD_PK) else 0
                                for key in columns.keys()
                            ]
                        )
                        == 1
                    ):
                        raise Exception(
                            error_msg.format(
                                "There is not exactly one primary key in the dict."
                            )
                        )
                    for key in columns.keys():
                        if columns[key].get(EC.QBC_FIELD_PK):
                            kwargs["primary_key_column_name"] = key
                            break
                chunking_column = chunking_column or kwargs["primary_key_column_name"]

            else:
                raise Exception(
                    "Arg chunking_interval must be integer or " + "datetime.timedelta!"
                )

        self.use_execution_date_for_incremental_loading = (
            use_execution_date_for_incremental_loading
        )
        self.timestamp_column = timestamp_column
        self.chunking_interval = chunking_interval
        self.chunking_column = chunking_column

        if kwargs.get("reload_data_from") and not chunking_interval:
            raise Exception(
                "When setting reload_data_from, must also set " + "chunking_interval!"
            )

        # run after setting class properties for templating
        super().__init__(*args, **kwargs)

        if (
            self.load_data_from
            or self.load_data_until
            or use_execution_date_for_incremental_loading
        ):
            if not timestamp_column:
                raise Exception(
                    "If you used data_from and/or data_until, you"
                    + " must also use timestamp_column to specify the column"
                    + " that is being used!"
                )

        # self.base_select is a SELECT statement (i.e. a string) ending in a
        #   WHERE {0} -> the extract process can add conditions!
        # self.base_sql is a pure SELECT statement ready to be executed
        if sql_select_statement:
            err_msg = "sql_select_statement and {0} cannot" " be used in combination!"
            if not where_clause == "1 = 1":
                raise Exception(err_msg.format("where_clause"))
            if source_table_name:
                raise Exception(err_msg.format("source_table_name"))
            self.base_sql = sql_select_statement
        else:
            source_table_name = source_table_name or target_table_name

            if self.columns_definition:
                self.base_sql = self._SQL_BASE.format(
                    **{
                        "columns": (
                            "\t"
                            + self._SQL_COLUMN_QUOTE
                            + (
                                "{0}\n,\t{0}".format(self._SQL_COLUMN_QUOTE).join(
                                    self.columns_definition.keys()
                                )
                            )
                            + self._SQL_COLUMN_QUOTE
                        ),
                        "database": source_database_name,
                        "schema": source_schema_name,
                        "table": source_table_name,
                        "where_clause": where_clause,
                    }
                )
            else:
                self.base_sql = self._SQL_BASE.format(
                    **{
                        "columns": "\t*",
                        "database": source_database_name,
                        "schema": source_schema_name,
                        "table": source_table_name,
                        "where_clause": where_clause,
                    }
                )
        self.base_select = self._SQL_BASE_SELECT.format(
            **{
                "select_sql": self.base_sql,
            }
        )

    def ewah_execute(self, context):
        str_format = "%Y-%m-%dT%H:%M:%SZ"

        if self.extract_strategy == EC.ES_FULL_REFRESH:
            self.log.info("Loading data as full refresh.")
        else:
            self.log.info(
                "Incrementally loading data from {0} to {1}.".format(
                    self.data_from.strftime(str_format),
                    self.data_until.strftime(str_format),
                )
            )

        params = {}
        sql_base = self.base_select
        if self.data_from and self.timestamp_column:
            sql_base = sql_base.format(
                "{0} >= {1} AND {{0}}".format(
                    self.timestamp_column,
                    self._SQL_PARAMS.format("data_from"),
                )
            )
            params["data_from"] = self.data_from
        if self.data_until and self.timestamp_column:
            sql_base = sql_base.format(
                "{0} <= {1} AND {{0}}".format(
                    self.timestamp_column,
                    self._SQL_PARAMS.format("data_until"),
                )
            )
            params["data_until"] = self.data_until
        sql_base = sql_base.format("1 = 1 {0}")

        chunking_column = self.chunking_column or self.timestamp_column
        if self.chunking_interval and chunking_column:

            previous_chunk, max_chunk = self._get_data_from_sql(
                sql=self._SQL_MINMAX_CHUNKS.format(
                    column=chunking_column,
                    base=sql_base.format(""),
                ),
                params=params,
                return_dict=False,
            )[0]
            if chunking_column == self.timestamp_column:
                tz = timezone("UTC")
                if previous_chunk and not previous_chunk.tzinfo:
                    previous_chunk = tz.localize(previous_chunk)
                if max_chunk and not max_chunk.tzinfo:
                    max_chunk = tz.localize(max_chunk)

            if previous_chunk is None or max_chunk is None:
                self.log.info("There appears to be no data?")
                return

            while previous_chunk <= max_chunk:
                params.update(
                    {
                        "from_value": previous_chunk,
                        "until_value": min(
                            max_chunk,
                            previous_chunk + self.chunking_interval,
                        ),
                    }
                )
                data = self._get_data_from_sql(
                    sql=sql_base.format(self._SQL_CHUNKING_CLAUSE).format(
                        column=chunking_column,
                        equal_sign=(
                            "="
                            if max_chunk < (previous_chunk + self.chunking_interval)
                            else ""
                        ),
                    ),
                    params=params,
                    return_dict=True,
                )

                self.upload_data(data=data)
                previous_chunk += self.chunking_interval
        else:
            params = params or None  # params must not be an empty dict, rather be None!
            self.upload_data(
                data=self._get_data_from_sql(
                    sql=sql_base.format("AND 1 = 1"),
                    return_dict=True,
                    params=params,
                ),
            )



from ewah.operators.sql_base import EWAHSQLBaseOperator
from ewah.hooks.postgres import EWAHPostgresHook


class EWAHPostgresOperator(EWAHSQLBaseOperator):

    _NAMES = ["pgsql", "postgres", "postgresql"]

    _SQL_BASE = 'SELECT\n{columns}\nFROM "{schema}"."{table}"\nWHERE {where_clause}'
    _SQL_BASE_SELECT = (
        "WITH raw_data AS ({select_sql}) SELECT * FROM raw_data WHERE {{0}}"
    )
    _SQL_COLUMN_QUOTE = '"'
    _SQL_MINMAX_CHUNKS = """
        WITH base AS ({base})
        SELECT MIN({column}), MAX({column})
        FROM base;
    """
    _SQL_CHUNKING_CLAUSE = """
        AND {column} >= %(from_value)s
        AND {column} <{equal_sign} %(until_value)s
    """
    _SQL_PARAMS = "%({0})s"

    _CONN_TYPE = EWAHPostgresHook.conn_type

    def _get_data_from_sql(self, sql, params=None, return_dict=True):
        self.log.info(
            "Executing:\n{0}\n\nWith params:\n{1}".format(
                sql,
                str(params),
            )
        )
        data = self.source_hook.get_data_from_sql(
            sql=sql,
            params=params,
            return_dict=return_dict,
        )
        return data



from ewah.constants import EWAHConstants as EC
from ewah.hooks.base import EWAHBaseHook

from psycopg2 import connect as pg_connect
from psycopg2.extras import RealDictCursor
from typing import Optional, List, Union, Dict, Any


class EWAHPostgresHook(EWAHBaseHook):

    _ATTR_RELABEL: dict = {
        "database": "schema",
        "hostname": "host",
        "user": "login",
    }

    conn_name_attr = "ewah_postgres_conn_id"
    default_conn_name = "ewah_postgres_default"
    conn_type = "ewah_postgres"
    hook_name = "EWAH PostgreSQL Connection"

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["extra"],
            "relabeling": {
                "password": "Password",
                "login": "User",
                "schema": "Database",
                "host": "Hostname / IP",
                "post": "Port (default: 5432)",
            },
        }

    @property
    def pgconn(self):
        if not hasattr(self, "_pgconn"):
            self._pgconn = pg_connect(
                "dbname='{0}' user='{1}' host='{2}' password='{3}' port='{4}'".format(
                    self.conn.database,
                    self.conn.user,
                    self.conn.host,
                    self.conn.password,
                    self.conn.port or "5432",
                )
            )
        return self._pgconn

    @property
    def cursor(self):
        """Cursor that returns lists of lists from the data source."""
        if not hasattr(self, "_cur"):
            self._cur = self.pgconn.cursor()
        return self._cur

    @property
    def dictcursor(self):
        """Cursor that returns lists of dictionaries from the data source."""
        if not hasattr(self, "_dictcur"):
            self._dictcur = self.pgconn.cursor(cursor_factory=RealDictCursor)
        return self._dictcur

    def get_records(self, sql, parameters=None):
        """
        Variant of execute method. Required to work with the SQL sensor.
        """
        return self.execute_and_return_result(
            sql=sql, params=parameters, return_dict=False
        )

    def execute(
        self, sql: str, params: Optional[dict] = None, commit: bool = False, cursor=None
    ) -> None:
        self.log.info(f"Executing SQL:\n\n{sql}\n\n")
        (cursor or self.cursor).execute(sql.strip(), vars=params)
        if commit:
            self.commit()

    def execute_and_return_result(
        self,
        sql: str,
        params: Optional[Dict[str, Any]] = None,
        return_dict: bool = False,
    ) -> Union[List[list], List[dict]]:
        cursor = self.dictcursor if return_dict else self.cursor
        self.execute(sql=sql, params=params, commit=False, cursor=cursor)
        return cursor.fetchall()

    def commit(self):
        self.log.info("Committing changes!")
        return self.pgconn.commit()

    def rollback(self):
        self.log.info("Rolling back changes!")
        return self.pgconn.rollback()

    def get_data_from_sql(
        self, sql: str, params: Optional[dict] = None, return_dict: bool = True
    ) -> Union[List[list], List[dict]]:
        cur = self.dictcursor if return_dict else self.cursor
        cur.execute(sql, vars=params)
        return cur.fetchall()

    def close(self):
        if hasattr(self, "_cur"):
            self._cur.close()
            del self._cur
        if hasattr(self, "_dictcur"):
            self._dictcur.close()
            del self._dictcur
        if hasattr(self, "_pgconn"):
            self.pgconn.close()
            del self._pgconn

    def __del__(self):
        self.close()



from ewah.hooks.base import EWAHBaseHook


class EWAHSSHHook(EWAHBaseHook):

    conn_name_attr = "ssh_conn_id"
    default_conn_name = "ssh_default"
    conn_type = "ewah_ssh"
    hook_name = "EWAH SSH Connection"

    @staticmethod
    def get_connection_form_widgets():
        """Returns connection widgets to add to connection form"""
        from wtforms import StringField
        from ewah.ewah_utils.widgets import EWAHTextAreaWidget

        return {
            "extra__ewah_ssh__private_key": StringField(
                "Private SSH Key (RSA Format)",
                widget=EWAHTextAreaWidget(rows=12),
            ),
        }

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["schema", "extra"],
            "relabeling": {
                "host": "SSH Host",
                "port": "SSH Port (defaults to 22)",
                "login": "SSH Username",
                "password": "SSH Password OR Private Key Password",
            },
        }
